use modules::EventLoop;

use toml::Value;
use toml::value::Table;

use hlua;
use hlua::{Lua, LuaFunction, AnyHashableLuaValue, AnyLuaValue, AsMutLua};

use {User, Message, Channel, NullMessageSender};
use event::{Event, Envelope};

use std::sync::Arc;
use std::sync::mpsc::Receiver;
use transformable_channels::mpsc::ExtSender;

use std::path::Path;
use std::fs::File;

use std::collections::HashMap;

pub struct LuaModule {
    code: Option<String>,
    file: Option<String>,
    variables: Table
}

impl LuaModule {
    pub fn new (_: &Table, config: &Table) -> Box<EventLoop> {
        Box::new(LuaModule {
            code: config.get("code").and_then(|value| value.as_str()).map(String::from),
            file: config.get("file").and_then(|value| value.as_str()).map(String::from),
            variables: config.clone()
        })
    }
}

struct TomlValueWrapper(Value);
implement_lua_read!(TomlValueWrapper);

impl TomlValueWrapper {
    pub fn lua_set (self, lua: &mut Lua, key: &str) {
        match self.0 {
            Value::String(string) => lua.set(key, string),
            Value::Integer(integer) => lua.set(key, (integer as i32)),
            Value::Float(float) => lua.set(key, float),
            Value::Boolean(boolean) => lua.set(key, boolean),
            Value::Table(table) => {},
            Value::Array(array) => {},
            Value::Datetime(datetime) => lua.set(key, datetime.to_string())
        }
    }
}

/*impl<'lua, L: AsMutLua<'lua>> hlua::Push<L> for TomlValueWrapper {
    type Err = hlua::Void;

    fn push_to_lua (self, lua: L) -> Result<hlua::PushGuard<L>, (hlua::Void, L)> {
        match self.0 {
            Value::String(string) => string.push_to_lua(lua),
            Value::Integer(integer) => (integer as i32).push_to_lua(lua),
            Value::Float(float) => float.push_to_lua(lua),
            Value::Boolean(boolean) => boolean.push_to_lua(lua),
            Value::Table(table) => {
                let hashmap: HashMap<_, _> = table.into_iter().map(|(key, value)| {
                    (key, TomlValueWrapper(value))
                }).collect();
                hashmap.push_to_lua(lua)
            },
            Value::Array(array) => {
                let vec: Vec<_> = array.into_iter().map(TomlValueWrapper).collect();
                vec.push_to_lua(lua)
            },
            Value::Datetime(datetime) => datetime.to_string().push_to_lua(lua)
        }
    }
}

impl<'lua, L: AsMutLua<'lua>> hlua::PushOne<L> for TomlValueWrapper {}*/

struct MessageWrapper {
    envelope: Arc<Envelope>
}

implement_lua_read!(MessageWrapper);
implement_lua_push!(MessageWrapper, |mut metatable| {
    let mut index = metatable.empty_array("__index");
    index.set("content", hlua::function1(|wrapper: &MessageWrapper| {
        if let Event::Message { ref message } = wrapper.envelope.event {
            Some(message.content.clone())
        } else {
            None
        }
    }));
    index.set("reply", hlua::function2(|wrapper: &MessageWrapper, reply: String| {
        if let Event::Message { ref message } = wrapper.envelope.event {
            message.reply(&reply);
        }
    }));
});

struct SenderWrapper {
    sender: Box<ExtSender<Event>>
}

implement_lua_read!(SenderWrapper);
implement_lua_push!(SenderWrapper, |mut metatable| {
    let mut index = metatable.empty_array("__index");
    index.set("send", hlua::function2(|wrapper: &SenderWrapper, data: HashMap<AnyHashableLuaValue, AnyLuaValue>| {
        if let Some(&AnyLuaValue::LuaString(ref event_type)) = data.get(&AnyHashableLuaValue::LuaString("type".to_owned())) {
            match event_type.as_ref() {
                "message" => {
                    wrapper.sender.send(Event::Message {
                        message: Message {
                            author: User {
                                name: data.get(&AnyHashableLuaValue::LuaString("username".to_owned()))
                                             .and_then(|value| match value {
                                                 &AnyLuaValue::LuaString (ref string_value) => Some(string_value.to_owned()),
                                                 _ => None
                                             })
                                             .unwrap_or_else(|| "".to_owned()),
                                sender: Box::new(NullMessageSender {})
                            },
                            channel: data.get(&AnyHashableLuaValue::LuaString("channel".to_owned()))
                                         .and_then(|value| match value {
                                             &AnyLuaValue::LuaString (ref string_value) => Some(string_value.to_owned()),
                                             &AnyLuaValue::LuaNumber (ref number_value) => Some(format!("{}", number_value)),
                                             _ => None
                                         }).map(|channel| Channel {
                                             name: channel,
                                             description: "".to_owned(),
                                             sender: Box::new(NullMessageSender {}),
                                             topic: "".to_owned()
                                         }),
                            content: data.get(&AnyHashableLuaValue::LuaString("message".to_owned()))
                                         .and_then(|value| match value {
                                             &AnyLuaValue::LuaString (ref string_value) => Some(string_value.to_owned()),
                                             _ => None
                                         })
                                         .expect("Invalid message event passed in")
                        }
                    });
                },
                _ => {}
            }
        }
    }));
});

impl EventLoop for LuaModule {
    fn run (&self, sender: Box<ExtSender<Event>>, receiver: Receiver<Arc<Envelope>>) {
        let mut lua = Lua::new();
        lua.openlibs();

        for (key, value) in &self.variables {
            if key != "type" {
                TomlValueWrapper(value.clone()).lua_set(&mut lua, key);
            }
        }

        lua.set("sender", SenderWrapper {
            sender: Box::new(sender.clone())
        });

        if let Some(ref file) = self.file {
            lua.execute_from_reader::<(), _>(File::open(&Path::new(file)).unwrap()).unwrap();
        }

        if let Some(ref code) = self.code {
            lua.execute(&code).unwrap()
        }

        loop {
            match receiver.recv() {
                Ok(envelope) => {
                    match envelope.event {
                        Event::Configure { ref configuration } => {
                            for (key, value) in configuration {
                                if key != "type" {
                                    TomlValueWrapper(value.clone()).lua_set(&mut lua, key);
                                }
                            }
                        },
                        Event::Message { ref message } => {
                            let on_message: Option<LuaFunction<_>> = lua.get("on_message");
                            match on_message {
                                Some(mut on_message) => {
                                    on_message.call_with_args::<(), _, _>(MessageWrapper {
                                        envelope: envelope.clone()
                                    }).unwrap();
                                },
                                None => {}
                            }
                        }
                        _ => ()
                    }
                }
                Err(error) => { error!("Error {:?}", error) }
            }
        }
    }
}
